using System;
using System.Collections.Generic;
using System.Text;
using System.Messaging;
using System.Xml.Serialization;
using NServiceBus.Unicast.Transport;
using Utils;
using NServiceBus.Messages;
using System.Threading;
using System.Transactions;

namespace NServiceBus.Unicast.Transport.Msmq
{
    public class MsmqTransport : ITransport, IDisposable
    {
        #region config info

        public string InputQueue
        {
            set
            {
                MessageQueue q = new MessageQueue(value);

                if (!QueueIsLocal(value))
                    throw new ApplicationException("Cannot work transactionally with remote queue.");
                else
                    SetLocalQueue(q);

                this.queue.Formatter = new BinaryMessageFormatter();
            }
        }

        public string ErrorQueue
        {
            set
            {
                this.errorQueue = new MessageQueue(value);
            }
        }

        private bool useXmlSerialization;
        public bool UseXmlSerialization
        {
            set { useXmlSerialization = value; }
        }

        private bool isTransactional;
        public bool IsTransactional
        {
            set { this.isTransactional = value; }
        }

        private bool skipDeserialization;
        public bool SkipDeserialization
        {
            set { skipDeserialization = value; }
        }

        private bool purgeOnStartup = false;
        public bool PurgeOnStartup
        {
            set { purgeOnStartup = value; }
        }

        private int numberOfWorkerThreads;
        public int NumberOfWorkerThreads
        {
            set { numberOfWorkerThreads = value; }
        }

        private TimeSpan timeToWaitForReceive = TimeSpan.FromSeconds(5);
        public string TimeToWaitForReceive
        {
            set { timeToWaitForReceive = TimeSpan.Parse(value); }
        }

        private string distributorControlAddress;
        public string DistributorControlAddress
        {
            set { distributorControlAddress = value; }
        }

        #endregion

        #region ITransport Members

        public event EventHandler<MsgReceivedEventArgs> MsgReceived;

        public string Address
        {
            get
            {
                return this.queue.Path;
            }
        }

        private IList<Type> messageTypesToBeReceived;
        public IList<Type> MessageTypesToBeReceived
        {
            set
            {
                this.messageTypesToBeReceived = value;

                if (this.useXmlSerialization)
                    this.serializer = new XmlSerializer(typeof(object), this.GetExtraTypes(value));
            }
        }

        public void Start()
        {
            if (this.purgeOnStartup)
                this.queue.Purge();

            for (int i = 0; i < this.numberOfWorkerThreads; i++)
                this.workerThreads.Add(new WorkerThread(this.Receive));

            foreach (WorkerThread w in this.workerThreads)
                w.Start();
        }

        public void ReceiveMessageLater(Msg m)
        {
            this.Send(m, this.Address);
        }

        public void Send(Msg m, string destination)
        {
            string address = this.Resolve(destination);

            using (MessageQueue q = new MessageQueue(address, QueueAccessMode.Send))
            {
                q.Formatter = this.queue.Formatter;

                Message toSend = new Message();

                if (m.Body == null && m.BodyStream != null)
                    toSend.BodyStream = m.BodyStream;
                else
                    this.serializer.Serialize(toSend.BodyStream, new List<object>(m.Body));


                toSend.ResponseQueue = new MessageQueue(m.ReturnAddress);
                toSend.Label = m.WindowsIdentityName;

                if (m.CorrelationId != null)
                    toSend.CorrelationId = m.CorrelationId;

                TimeSpan timeToBeReceived = MessageQueue.InfiniteTimeout;
                if (m.Body != null)
                {
                    foreach (IMessage message in m.Body)
                        if (message.GetType().GetCustomAttributes(typeof(RecoverableAttribute), true).Length > 0)
                            toSend.Recoverable = true;

                    foreach (IMessage message in m.Body)
                        foreach (TimeToBeReceivedAttribute a in message.GetType().GetCustomAttributes(typeof(TimeToBeReceivedAttribute), true))
                            timeToBeReceived = (a.TimeToBeReceived < timeToBeReceived ? a.TimeToBeReceived : timeToBeReceived);
                }

                if (timeToBeReceived < MessageQueue.InfiniteTimeout)
                    toSend.TimeToBeReceived = timeToBeReceived;

                q.Send(toSend, this.GetTransactionTypeForSend());

                m.Id = toSend.Id;
            }
        }

        #endregion

        #region helper methods

        private void Receive()
        {
            if (this.isTransactional)
                new TransactionWrapper().RunInTransaction(new Callback(this.ReceiveFromQueue));
            else
                this.ReceiveFromQueue();
        }

        public void ReceiveFromQueue()
        {
            try
            {
                if (this.distributorControlAddress != null)
                    if (!sentReadyMessage)
                    {
                        Msg ready = new Msg();
                        ready.Body = new IMessage[] { new ReadyMessage() };
                        ready.ReturnAddress = this.Address;
                        ready.WindowsIdentityName = Thread.CurrentPrincipal.Identity.Name;

                        this.Send(ready, this.distributorControlAddress);
                        sentReadyMessage = true;
                    }

                Message m = this.queue.Receive(this.timeToWaitForReceive, this.GetTransactionTypeForReceive());

                Msg result = Convert(m);

                try
                {
                    if (this.skipDeserialization)
                        result.BodyStream = m.BodyStream;
                    else
                        result.Body = this.Extract(m);
                }
                catch
                {
                    if (this.errorQueue != null)
                        this.errorQueue.Send(m, this.GetTransactionTypeForSend());

                    return;
                }

                if (this.MsgReceived != null)
                    this.MsgReceived(this, new MsgReceivedEventArgs(result));
            }
            catch (MessageQueueException mqe)
            {
                if (mqe.MessageQueueErrorCode == MessageQueueErrorCode.IOTimeout)
                    return;
            }

            sentReadyMessage = false;
            return;
        }

        public static bool QueueIsLocal(string value)
        {
            string machineName = Environment.MachineName.ToLower();

            value = value.Replace("FormatName:DIRECT=OS:", "");
            int index = value.IndexOf('\\');

            string queueMachineName = value.Substring(0, index).ToLower();

            return (machineName == queueMachineName || queueMachineName == "localhost" || queueMachineName == ".");
        }

        public static Msg Convert(Message m)
        {
            Msg result = new Msg();
            result.Id = m.Id;
            result.CorrelationId = m.CorrelationId;

            if (m.ResponseQueue != null)
                result.ReturnAddress = m.ResponseQueue.Path;

            result.WindowsIdentityName = m.Label;

            return result;
        }

        private string Resolve(string destination)
        {
            string dest = destination.ToUpper().Replace("FORMATNAME:DIRECT=OS:", "");

            MessageQueue q = new MessageQueue(dest);
            if (QueueIsLocal(dest))
                q.MachineName = Environment.MachineName;

            return "FormatName:DIRECT=OS:" + q.Path;
        }

        private IMessage[] Extract(Message message)
        {
            IMessage[] result;

            if (this.serializer != null)
            {
                try
                {
                    List<object> body = serializer.Deserialize(message.BodyStream) as List<object>;

                    result = new IMessage[body.Count];

                    int i = 0;
                    foreach (IMessage m in body)
                        result[i++] = m;
                }
                catch
                {
                    return null;
                }
            }
            else
                result = message.Body as IMessage[];

            return result;
        }

        private MessageQueueTransactionType GetTransactionTypeForReceive()
        {
            if (this.isTransactional)
                return MessageQueueTransactionType.Automatic;
            else
                return MessageQueueTransactionType.None;
        }

        private MessageQueueTransactionType GetTransactionTypeForSend()
        {
            if (this.isTransactional)
            {
                if (Transaction.Current != null)
                    return MessageQueueTransactionType.Automatic;
                else
                    return MessageQueueTransactionType.Single;
            }
            else
                return MessageQueueTransactionType.Single;
        }

        private void SetLocalQueue(MessageQueue q)
        {
            q.MachineName = Environment.MachineName; // just in case we were given "localhost"

            if (!q.Transactional)
                throw new ArgumentException("Queue must be transactional (" + q.Path + ").");
            else
                this.queue = q;

            MessagePropertyFilter mpf = new MessagePropertyFilter();
            mpf.SetAll();

            this.queue.MessageReadPropertyFilter = mpf;
        }

        private Type[] GetExtraTypes(IList<Type> value)
        {
            List<Type> types = new List<Type>(value);
            foreach (Type t in value)
                if (this.CantSerializeType(t))
                    types.Remove(t);

            if (!types.Contains(typeof(List<object>)))
                types.Add(typeof(List<object>));

            return types.ToArray();
        }

        private bool CantSerializeType(Type t)
        {
            if (t.IsInterface || t.IsAbstract)
                return true;

            if (t.IsArray)
                return CantSerializeType(t.GetElementType());

            return false;
        }

        #endregion

        #region members

        private MessageQueue queue;
        private MessageQueue errorQueue;
        private XmlSerializer serializer = null;
        IList<WorkerThread> workerThreads = new List<WorkerThread>();

        /// <summary>
        /// ThreadStatic
        /// </summary>
        [ThreadStatic]
        static bool sentReadyMessage;

        #endregion

        #region IDisposable Members

        public void Dispose()
        {
            this.queue.Dispose();
        }

        #endregion
    }
}
